
# include <math.h>
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include "mppath.h"
# include "mpmp.h" /* to access internals of MP, like jump_buf and arith_error */

#define scaled_from_double(a) (int)((a)*65536.0)
#define double_from_scaled(a) (double)((a)/65536.0)
#define PI 3.1415926535897932384626433832795028841971 

#define PATH_LOC(A) mp->mem[mp->mem[(A+7)].hh.rh + 1].hh.rh
#define TOO_LARGE(a) (fabs((a))>=4096.0)
#define ROUNDED_ZERO(v) (fabs((v))<0.00001 ? 0 : (v))

extern int mp_gr_unexport(MP mp, struct mp_edge_object *hh);
extern void mp_make_choices (MP mp, int knots);
extern struct mp_edge_object *mp_gr_export(MP mp, int h);
extern void mp_flush_cur_exp (MP mp,int v) ;

#define left_curl left_x /* curl or given information when entering this knot */
#define left_tension left_y /* tension information when entering this knot */
#define right_curl right_x /* curl or given information when leaving this knot */
#define right_tension right_y /* tension information when leaving this knot */


static mp_edge_object * create_edges (MP mp)
{
    mp_edge_object *h = (mp_edge_object *)malloc(sizeof(mp_edge_object));
    if (h==NULL) return NULL;
    memset(h,0,sizeof(mp_edge_object));
    h->body = (mp_graphic_object *)malloc(sizeof(mp_fill_object));
    if (h->body==NULL) return NULL;
    memset(h->body,0,sizeof(mp_fill_object));
    h->body->type = mp_fill_code;
    return h;
}

static mp_knot *copy_knot (mp_knot *p) 
{
  mp_knot *q; /* the copy */
  if (p==NULL) 
    return NULL;
  q = malloc(sizeof (mp_knot));
  if (q == NULL) return NULL;
  memcpy(q,p,sizeof (mp_knot));
  q->next=NULL;
  return q;
}


static mp_knot *copy_path (mp_knot *p) {
  mp_knot *q, *pp, *qq; /* for list manipulation */
  if (p==NULL) 
    return NULL;
  q=copy_knot(p);
  if (q==NULL) return NULL;
  qq=q; 
  pp=p->next;
  while ( pp!=p ) { 
    qq->next=copy_knot(pp);
    if (qq->next==NULL) return NULL;
    qq=qq->next;
    pp=pp->next;
    if (pp==NULL) return NULL;
  }
  qq->next=q;
  return q;
}

static int clone_path(mp_knot *source,mp_knot *target)
{
    mp_knot *sourcehead;
    if (source==NULL || target==NULL) return 0;
    sourcehead = source;
    do {
	mp_knot *next = target->next;
	memcpy(target,source,sizeof(mp_knot));
	target->next = next;
	source = source->next;
	target = target->next;
    } while (source && source != sourcehead);
    return (source!=NULL);
}

static int path_needs_fixing (mp_knot *source) {
    mp_knot *sourcehead;
    if (source==NULL) return 0;
    sourcehead = source;
    do {
	source = source->next;
    } while (source && source != sourcehead);
    if (!source) {
	return 1;
    }
    return 0;
}


int mp_cycle_knotpair (MP mp, mp_knot *p, mp_knot *q)
{
    if (p==NULL ||q==NULL) return 0;
    p->next = q;
    p->right_tension = scaled_from_double(1.0);
    if (p->right_type==mp_endpoint) {
	p->right_type = mp_open;
    }
    q->left_tension = scaled_from_double(1.0);
    if (q->left_type == mp_endpoint) {
	q->left_type = mp_open;
    }
    return 1;
}

int mp_final_knotpair (MP mp, mp_knot *q, mp_knot *first)
{
    if (q==NULL || first==NULL) return 0;
    q->next = first;
    q->right_type = mp_endpoint;
    q->right_tension = scaled_from_double(1.0);
    first->left_type = mp_endpoint;
    first->left_tension = scaled_from_double(1.0);
    return 1;
}

mp_knot * mp_create_knot (MP mp)
{
    mp_knot *q = (mp_knot *)malloc(sizeof(mp_knot));
    if (q==NULL) return NULL;
    memset(q,0,sizeof(mp_knot));
    q->left_type = mp_endpoint;
    q->right_type = mp_endpoint;
    return q;
}

int mp_set_knot (MP mp, mp_knot *p, double x, double y)
{
    if (TOO_LARGE(x)) return 0;
    if (TOO_LARGE(y)) return 0;
    if (p==NULL) return 0;
    p->x_coord = scaled_from_double(x);
    p->y_coord = scaled_from_double(y);
    return 1;
}

mp_knot * mp_append_knot (MP mp, mp_knot *p, double x, double y)
{
    mp_knot *q = mp_create_knot(mp);
    if (q==NULL) return NULL;
    if (!mp_set_knot(mp, q, x, y)) {
	free(q);
	return NULL;
    }
    if (p == NULL) return q;
    if (!mp_cycle_knotpair(mp, p,q)) {
	free(q);
	return NULL;
    }
    return q;
}

int mp_set_knot_curl (MP mp, mp_knot *q, double value) {
    if (q==NULL) return 0;
    if (TOO_LARGE(value)) return 0;
    q->right_type=mp_curl; 
    q->right_curl= scaled_from_double(value);
    if (q->left_type==mp_open) {
	q->left_type=mp_curl; 
	q->left_curl= scaled_from_double(value);
    }
    return 1;
}

int mp_set_knotpair_curls (MP mp, mp_knot *p, mp_knot *q, double t1, double t2) {
    if (p==NULL || q==NULL) return 0;
    if (mp_set_knot_curl(mp, p, t1))
	return mp_set_knot_curl(mp, q, t2);
    return 0;
}

int mp_set_knotpair_tensions (MP mp, mp_knot *p, mp_knot *q, double t1, double t2) {
    if (p==NULL || q==NULL) return 0;
    if (TOO_LARGE(t1)) return 0;
    if (TOO_LARGE(t2)) return 0;
    if ((fabs(t1)<0.75)) return 0;
    if ((fabs(t2)<0.75)) return 0;
    p->right_tension = scaled_from_double(t1);
    q->left_tension = scaled_from_double(t2);
    return 1;
}

int mp_set_knotpair_controls (MP mp, mp_knot *p, mp_knot *q, double x1, double y1, double x2, double y2) {
    if (p==NULL || q==NULL) return 0;
    if (TOO_LARGE(x1)) return 0;
    if (TOO_LARGE(y1)) return 0;
    if (TOO_LARGE(x2)) return 0;
    if (TOO_LARGE(y2)) return 0;
    p->right_type=mp_explicit; 
    p->right_x= scaled_from_double(x1);
    p->right_y= scaled_from_double(y1);
    q->left_type=mp_explicit; 
    q->left_x= scaled_from_double(x2);
    q->left_y= scaled_from_double(y2);
    return 1;
}

int mp_set_knot_direction (MP mp, mp_knot *q, double x, double y) {
    if (q==NULL) return 0;
    double value = 0;
    if (TOO_LARGE(x)) return 0;
    if (TOO_LARGE(y)) return 0;
    if (!(x==0 && y == 0))
	value = atan2 (y, x) * (180.0 / PI)  * 16.0;
    q->right_type=mp_given; 
    q->right_curl= scaled_from_double(value);
    if (q->left_type==mp_open) {
	q->left_type=mp_given; 
	q->left_curl= scaled_from_double(value);
    }
    return 1;
}


int mp_set_knotpair_directions (MP mp, mp_knot *p, mp_knot *q, double x1, double y1, double x2, double y2) {
    if (p==NULL || q==NULL) return 0;
    if (mp_set_knot_direction(mp,p, x1, y1))
	return mp_set_knot_direction(mp,q, x2, y2);
    return 0;
}

int mp_solve_path (MP mp, mp_knot *first)
{
    int cur_exp;
    int saved_arith_error;
    jmp_buf *saved_jump_buf;
    mp_fill_object *f;
    mp_edge_object *h;
    if (first==NULL) return 0;
    h = create_edges(mp);
    if (h==NULL) return 0;
    f = (mp_fill_object *)h->body;
    if (path_needs_fixing(first)) return 0;
    f->path_p = copy_path(first);
    if (f->path_p==NULL) return 0;
    saved_jump_buf = mp->jump_buf;
    mp->jump_buf = malloc(sizeof(jmp_buf));
    if (mp->jump_buf == NULL || setjmp(*(mp->jump_buf)) != 0) {   
       return 0; 
    }    
    cur_exp =  mp_gr_unexport(mp, h); /* this deallocs h for us */
    saved_arith_error = mp->arith_error;
    mp->arith_error=0;
    mp_make_choices(mp, PATH_LOC(cur_exp));
    if (mp->arith_error) {
	free(mp->jump_buf);
	mp->jump_buf = saved_jump_buf;
	mp_flush_cur_exp(mp, 0);  /* kill off cur_exp */
 	return 0;
    }
    mp->arith_error = saved_arith_error;
    h = mp_gr_export(mp, cur_exp); /* aborts on malloc failure */
    mp_flush_cur_exp(mp, 0); /* kill off cur_exp */
    f = (mp_fill_object *)h->body;
    if (!clone_path(f->path_p, first)) {
	free(mp->jump_buf);
	mp->jump_buf = saved_jump_buf;
	mp_gr_toss_objects(h);
 	return 0;
    }
    free(mp->jump_buf);
    mp->jump_buf = saved_jump_buf;
    mp_gr_toss_objects(h);
    return 1;
}

void mp_dump_path (MP mp, mp_knot *h) {
    mp_knot *p, *q;
    if (h == NULL) return;
    p = h;
    do {  
	q=p->next;
	if ( (p==NULL)||(q==NULL) ) { 
	    printf("\n???"); 
	    return; /* this won't happen */
	}
	printf ("(%g,%g)", double_from_scaled(p->x_coord), double_from_scaled(p->y_coord));
	switch (p->right_type) {
	case mp_endpoint: 
	    if ( p->left_type==mp_open ) printf("{open?}"); 
	    if ( (q->left_type!=mp_endpoint)||(q!=h) ) 
		q=NULL; /* force an error */
	    goto DONE1;
	    break;
	case mp_explicit: 
	    printf ("..controls (%g,%g)", double_from_scaled(p->right_x), double_from_scaled(p->right_y));
	    printf(" and ");
	    if ( q->left_type!=mp_explicit ) { 
		printf("??"); /* can't happen */
	    } else {
		printf ("(%g,%g)",double_from_scaled(q->left_x), double_from_scaled(q->left_y));
	    }
	    goto DONE1;
	    break;
	case mp_open: 
	    if ( (p->left_type!=mp_explicit)&&(p->left_type!=mp_open) ) {
		printf("{open?}"); /* can't happen */
	    }
	    break;
	case mp_curl:
	case mp_given: 
	    if ( p->left_type==mp_open )  
		printf("??"); /* can't happen */
	    if ( p->right_type==mp_curl ) { 
		printf("{curl %g}", double_from_scaled(p->right_curl));
	    } else { 
		double rad = (double_from_scaled(p->right_curl) / 16.0) * PI/180.0;
		double n_cos = ROUNDED_ZERO(cos(rad)*4096);
		double n_sin = ROUNDED_ZERO(sin(rad)*4096);
		printf("{%g,%g}", n_cos, n_sin);
	    }
	    break;
	default:
	    printf("???"); /* can't happen */
	    break;
	}
	if ( q->left_type<=mp_explicit ) {
	    printf("..control?"); /* can't happen */
	} else if ( (p->right_tension!=scaled_from_double(1.0))||
		    (q->left_tension!=scaled_from_double(1.0)) ) {
	    printf("..tension ");
	    if ( p->right_tension<0 ) 
		printf("atleast ");
	    printf("%g",double_from_scaled(abs(p->right_tension)));
	    if ( p->right_tension!=q->left_tension ) { 
		printf(" and ");
		if ( q->left_tension<0 ) 
		    printf("atleast ");
		printf("%g",double_from_scaled(abs(q->left_tension)));
	    }
	}
    DONE1:
	p=q;
	if ( p!=h || h->left_type!=mp_endpoint) {
	    printf ("\n ..");
	    if ( p->left_type == mp_given ) { 
		double rad = (double_from_scaled(p->left_curl) / 16.0) * PI/180.0;
		double n_cos = ROUNDED_ZERO(cos(rad)*4096);
		double n_sin = ROUNDED_ZERO(sin(rad)*4096);
		printf("{%g,%g}", n_cos, n_sin);
	    } else if ( p->left_type ==mp_curl ){ 
		printf("{curl %g}", double_from_scaled(p->left_curl)); 
	    }
	}
    } while (p!=h);
    if ( h->left_type!=mp_endpoint ) 
	printf("cycle");
    printf (";\n");
}

void mp_free_path (MP mp, mp_knot *p) {
    mp_knot *q; /* the node being freed */
    mp_knot *r; /* the next node */
    if (p==NULL)
	return;
    q=p;
    do {  
	r=q->next; 
	free(q); q=r;
    } while (q && q!=p);
}
